import sys


class Progressbar:
    def __init__(self, iterator=None, total=None, length=80):
        self.it = iterator
        if total is None and iterator is not None:
            try:
                self.total = len(iterator)
            except Exception:
                self.total = None
        self.length = length

    def __iter__(self):
        self.elapsed = 0
        self.update()
        for x in self.it:
            yield x
            self.elapsed += 1
            self.update()

    def __repr__(self):
        if self.total is not None:
            p = int(self.length * self.elapsed / self.total)
            tot = self.total
        else:
            p = 0
            tot = "???"
        return "|" + "=" * p + " " * (self.length - p) + f"| ({self.elapsed}/{tot})"

    def update(self):
        sys.stderr.write(self.__repr__() + "\r")



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Handle elements to calculate mass, abundance, etc.
"""

import os
import re
import sqlite3

from .constants import me
from .misc import deprecated


def formulafy(x):
    """
    Convert the input string x to a Latex representation of the chemical formula.
    e.g. ^13CH4 -> $^{13}CH_{4}$
    This can be used in matplotlib to display nicely a chemical formula.
    """
    import re

    return (
        "$"
        + re.sub(
            "([a-zA-Z])_?([0-9]+)",
            r"\1_{\2}",
            re.sub(r"\^([0-9]+)", r"^{\1}", re.sub("([\\+-]+)$", r"^{\1}", x)),
        )
        + "$"
    )


def get_peaklist(nominal_mass, negative=False):
    """
    Retrieve all elements from the database which have a given nominal_mass

    Parameters
    ----------
    nominal_mass : int or tuple/list
        If int provide the nominal mass of the elements
        If tuple/list return all elements bounded between a lower (first element) and upper (second eleemnt) nominal mass.
    negative : boolean
        If True will return the elements in the database for negative polarity

    Examples
    --------
    get_peaklist(12) will return all elements with a nominal mass of 12u
    get_peaklist((12,15)) will return all elements with a nominal mass between 12 and 15
    """
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    if type(nominal_mass) in [float, int]:
        c.execute(
            "SELECT Formula from fragments where NominalMass={nm} and (Polarity is NULL or Polarity=={pol})".format(
                nm=nominal_mass, pol=[1, -1][negative]
            )
        )
    elif type(nominal_mass) in [tuple, list]:
        c.execute(
            "SELECT Formula from fragments where NominalMass>={nm[0]} and NominalMass<={nm[1]} and (Polarity is NULL or Polarity=={pol})".format(
                pol=[1, -1][negative]
            )
        )
    return [str(x[0]) for x in c.fetchall()]


def get_properties(elt):
    """
    Retrieve properties of an element such as its density, lattice constant, etc. as a dictionary
    If no information is found in the database then at least Element, Z and mass will be returned
    Examples
    --------
    > get_properties("C")
    {'Element': 'C', 'Z': 6, 'mass': 12.010735896764459}

    > get_properties("Si")
    {'Element': 'Si',
    'Q': 0.78,
    'Z': 14,
    'density': 5e+22,
    'ion_nrj': 8.15,
    'lattice_const': 5.431,
    'lattice_type': 'diamond_fcc',
    'mass': 28.085498706418875,
    'sublim_nrj': 4.63,
    'work_func': 4.85}
    """
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    c.execute(f"SELECT * from properties where Element='{elt}'")
    res = c.fetchone()
    r = {}
    if res:
        for idx, col in enumerate(c.description):
            r[col[0]] = res[idx]
    else:
        res = c.execute(
            f"SELECT mass,abund,symbol,Z from elements where symbol='{elt}'"
        )
        if res:
            row = res.fetchall()
            r["Element"] = row[0][2]
            r["Z"] = row[0][3]
            r["mass"] = sum([x[0] * x[1] for x in row])
    return r


def get_mass(elt, mz=True):
    """
    Calculate the exact molar mass of a givent chemical compound.

    Parameters
    ----------
    elt : string
        Element given by its chemical formula
    mz : boolean
        Will output the mass per ion.
        e.g. This means that if the formula is Si++, then the answer will be half the mass of Si

    Notes
    -----
    "Si+" will subtract the mass of one electron compared to "Si"
    "Si-" will add the mass of one electron compared to "Si"
    """
    if type(elt) is Molecule:
        elt = str(elt)
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    m = 0
    for A, x, n in re.findall("(?:\\^([0-9]+))?([A-Z][a-z]?)_?([0-9]*)", elt):
        n = 1 if n == "" else int(n)
        if A == "":
            c.execute(
                f"SELECT mass from elements where symbol='{x}' and abund=(SELECT max(abund) from elements where symbol='{x}')"
            )
        else:
            c.execute(f"SELECT mass from elements where symbol='{x}' and A={A}")
        res = c.fetchone()
        if res is None:
            raise Exception(f"Cannot fetch mass of {x}")
        m += n * res[0]
    m -= me * elt.count("+")
    m += me * elt.count("-")
    Z = max(1, abs(elt.count("+") - elt.count("-")))
    if mz:
        return m / Z
    return m


def get_main_isotope(elt):
    """
    Return the A value of the main isotope of a given element.
    e.g. get_main_isotope("Si") will return 28
    """
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    c.execute(
        f"SELECT A from elements where symbol='{elt}' and abund=(SELECT max(abund) from elements where symbol='{elt}')"
    )
    res = c.fetchone()
    if res is None:
        raise Exception(f"Cannot fetch element '{elt}'")
    return res[0]


def is_main_isotope(elt, A):
    return get_main_isotope(elt) == A


def get_isotopes_of_element(elt):
    """
    Get all the isotopes for a given element
    """
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    c.execute(f"SELECT symbol, A, abund from elements where symbol='{elt}'")
    return c.fetchall()


def _formula2dict(elt, iso=True):
    """
    Convert a formula into a dictionary.
    """
    elts = {}
    for A, x, n in re.findall("(?:\\^([0-9]+))?([A-Z][a-z]?)_?([0-9]*)", elt):
        A = get_main_isotope(x) if A == "" else int(A)
        n = 1 if n == "" else int(n)
        if iso:
            if (A, x) not in elts:
                elts[(A, x)] = n
            else:
                elts[(A, x)] += n
        else:
            if x not in elts:
                elts[x] = n
            else:
                elts[x] += n
    return elts


def is_fragment_of(fragment, element):
    f = _formula2dict(fragment, iso=False)
    e = _formula2dict(element, iso=False)
    return all(not (x not in e or e[x] < f[x]) for x in f)


def simplify_formula(elt, debug=False):
    elts = _formula2dict(elt)
    return _dict2formula(elts)


def _dict2formula(elts, debug=False):
    if debug:
        print(elts)
    order = "CHSNO"
    res = ""
    for o in order:
        filt = [x[0] for x in elts if x[1] == o]
        miso = get_main_isotope(o)
        filt2 = [x for x in filt if x != miso]
        for A in filt2:
            n = elts[(A, o)]
            if n == 0:
                continue
            fmt = "^{A}{sym}"
            if n > 1:
                fmt += "{n}"
            if debug:
                print("-", fmt, A, o, n)
            res += fmt.format(A=A, sym=o, n=n)
        if (miso, o) in elts:
            n = elts[(miso, o)]
            if n == 0:
                break
            fmt = "{sym}"
            if n > 1:
                fmt += "{n}"
            if debug:
                print("*", fmt, A, o, n)
            res += fmt.format(sym=o, n=n)
    for A, sym in sorted([x for x in elts if x[1] not in order], key=lambda x: x[1]):
        n = elts[(A, sym)]
        if n == 0:
            continue
        fmt = "{sym}" if is_main_isotope(sym, A) else "^{A}{sym}"
        if n > 1:
            fmt += "{n}"
        res += fmt.format(A=A, sym=sym, n=n)
    return res


def _get_isotopes_elt(n, x, min_abund=0):
    """
    get the isotopes of n atoms x
    """
    res = {"": 1}
    iso = get_isotopes_of_element(x)
    for _i in range(n):
        Nres = {}
        for x in res:
            for y in iso:
                X = simplify_formula(x + "^{1}{0}".format(*y))
                ab = res[x] * y[2]
                if X not in Nres:
                    if n * ab >= min_abund:
                        Nres[X] = ab
                else:
                    Nres[X] += ab
        res = Nres
    return [(x, res[x]) for x in res]


def get_isotopes(elt, min_abund=0):
    charges = "+" * elt.count("+") + "-" * elt.count("-")
    res = {"": 1}
    for _A, x, n in re.findall("(?:\\^([0-9]+))?([A-Z][a-z]?)_?([0-9]*)", elt):
        n = 1 if n == "" else int(n)
        R = _get_isotopes_elt(n, x, min_abund=min_abund)
        Nres = {}
        for x0 in res:
            for x, ab0 in R:
                ab = res[x0] * ab0
                if ab >= min_abund:
                    Nres[x0 + x] = ab
        res = Nres
    R = [(x + charges, get_mass(x + charges), res[x]) for x in res]
    return R


def get_abund(elt):
    from .math import perm, prod

    if type(elt) is Molecule:
        elt = str(elt)
    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    abund = 1
    elts = [
        (int([A, get_main_isotope(x)][A == ""]), x, int(n + ["", "1"][n == ""]))
        for A, x, n in re.findall("(?:\\^([0-9]+))?([A-Z][a-z]?)_?([0-9]*)", elt)
    ]
    ns = {}

    for x in elts:
        if x[1] not in ns:
            ns[x[1]] = []
        ns[x[1]].append(x[2])

    permutations = prod([perm(ns[k]) for k in ns])

    for A, x, n in elts:
        if A == "":
            c.execute(
                f"SELECT abund from elements where symbol='{x}' and abund=(SELECT max(abund) from elements where symbol='{x}')"
            )
        else:
            c.execute(f"SELECT abund from elements where symbol='{x}' and A={A}")
        res = c.fetchone()
        if res is None:
            raise Exception(f"Cannot fetch mass of {x}")
        abund *= (res[0]) ** n
    return permutations * abund


@deprecated("getOrganicAt")
def get_organic_at(m0):
    import numpy as np

    res = []
    for C in range(1, 1 + m0 // 12):
        for N in range(1 + (m0 - 12 * C) // 15):
            for S in range(1 + (m0 - 12 * C - 15 * N) // 32):
                for O in range(1 + (m0 - 12 * C - 15 * N - 32 * S) // 16):
                    H = m0 - 12 * C - 15 * N - 32 * S - 16 * O
                    elt = ""
                    for e in "CNSOH":
                        if locals()[e] > 0:
                            elt += e
                            if locals()[e] > 1:
                                elt += str(locals()[e])
                    if int(np.round(get_mass(elt), 0)) == m0:
                        res.append(elt)
    return res


def elts_substract(elt1, elt2):
    r = _formula2dict(elt1)
    d = _formula2dict(elt2)
    for x in d:
        if x not in r:
            raise Exception(
                "Cannot subtract a molecule with more atoms than the main one"
            )
        r[x] -= d[x]


def _dict_add(d1, d2):
    r = {k: d1[k] for k in d1}
    for k in d2:
        r[k] = r.get(k, 0) + d2[k]
    return r


def _elts_add(eltsa, eltsb):
    r = []
    for x, mx in eltsa:
        for y, my in eltsb:
            r.append((_dict_add(x, y), mx + my))
    return r


@deprecated("eltsNM")
def elts_nm(elts, NM):
    import re

    r = [({}, 0)]
    res = []
    if isinstance(elts, str):
        elts = re.compile(r"((?:^[0-9]+)?[A-Z][a-z]?(?:_[0-9]+)?)").findall(elts)
    me = [get_mass(x) for x in elts]
    while r:
        rnew = _elts_add(r, list(zip([_formula2dict(x) for x in elts], me)))
        r = [x for x in rnew if x[1] < NM + 0.5]
        res += [
            _dict2formula(x[0])
            for x in r
            if x[1] > NM - 0.5
            if _dict2formula(x[0]) not in res
        ]
        res = list(set(res))
    return res


class Molecule:
    def __init__(self, formula):
        self.atoms = _formula2dict(formula)

    def formula(self):
        return _dict2formula(self.atoms)

    def __add__(self, b):
        r = Molecule(self.formula())
        for x in b.atoms:
            if x in r.atoms:
                r.atoms[x] += b.atoms[x]
            else:
                r.atoms[x] = b.atoms[x]
        return r

    __radd__ = __add__

    def __mul__(self, n):
        assert isinstance(n, int)
        r = Molecule(self.formula())
        for x in r.atoms:
            r.atoms[x] *= n
        return r

    __rmul__ = __mul__

    def __sub__(self, b):
        r = Molecule(self.formula())
        for x in b.atoms:
            n = b.atoms[x]
            if x not in r.atoms or r.atoms[x] < n:
                raise Exception("Cannot subtract a larger molecule from a smaller one")
            r.atoms[x] -= n
        return r

    def inc(self, elt="H"):
        r = Molecule(self.formula())
        k = next(iter(_formula2dict(elt).keys()))
        if k in r.atoms:
            r.atoms[k] += 1
        else:
            r.atoms[k] = 1
        return r

    def dec(self, elt="H"):
        r = Molecule(self.formula())
        k = next(iter(_formula2dict(elt).keys()))
        if k not in r.atoms or r.atoms[k] < 1:
            raise Exception("Cannot decrement molecule below 0 atoms")
        r.atoms[k] -= 1
        return r

    def mass(self):
        return get_mass(self.formula())

    def abund(self):
        return get_abund(self.formula())

    def __repr__(self):
        return self.formula()

    def __str__(self):
        return self.formula()


H = Molecule("H")



# Form implementation generated from reading ui file 'timer_display.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtWidgets


class Ui_ToF_Timer:
    def setupUi(self, ToF_Timer):
        ToF_Timer.setObjectName("ToF_Timer")
        ToF_Timer.resize(466, 108)
        self.centralWidget = QtWidgets.QWidget(ToF_Timer)
        self.centralWidget.setObjectName("centralWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralWidget)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.progressBar = QtWidgets.QProgressBar(self.centralWidget)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout.addWidget(self.progressBar)
        self.label = QtWidgets.QLabel(self.centralWidget)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(self.centralWidget)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.label_3 = QtWidgets.QLabel(self.centralWidget)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        ToF_Timer.setCentralWidget(self.centralWidget)

        self.retranslateUi(ToF_Timer)
        QtCore.QMetaObject.connectSlotsByName(ToF_Timer)

    def retranslateUi(self, ToF_Timer):
        _translate = QtCore.QCoreApplication.translate
        ToF_Timer.setWindowTitle(_translate("ToF_Timer", "ToF_Timer"))
        self.label.setText(_translate("ToF_Timer", "Remaining time: 00:00:00"))
        self.label_2.setText(_translate("ToF_Timer", "Scans: 0/0"))
        self.label_3.setText(_translate("ToF_Timer", "Analysis time: 0 s"))



import os
import struct
import sys

import numpy as np
from PyQt5.QtCore import QDir, QFileInfo, QSettings, Qt
from PyQt5.QtWidgets import (
    QApplication,
    QFileDialog,
    QMainWindow,
    QShortcut,
    QTableWidgetItem,
)

import pySPM
from pySPM.tools.spectraviewer import Ui_SpectraViewer

DPI = 100.0


class SpectraViewer(QMainWindow):
    def __init__(self, filename=None, parent=None):
        super().__init__(parent)
        self.ui = Ui_SpectraViewer()
        self.ui.setupUi(self)
        self.sf = 7200
        self.k0 = 0
        self.dsf = 0
        self.dk0 = 0
        self.ita = None
        self.fig = self.ui.mpl.canvas.fig
        self.canvas = self.ui.mpl.canvas
        self.ax = self.fig.add_subplot(111)
        self.nextMass = QShortcut(Qt.Key_Plus, self)
        self.prevMass = QShortcut(Qt.Key_Minus, self)
        self.nextMass.activated.connect(self.next_mass)
        self.prevMass.activated.connect(self.prev_mass)
        self.ui.pushButton_2.clicked.connect(self.toggleMassCal)
        self.ui.pushButton.clicked.connect(self.removeMassCalItem)
        self.ui.show_mass.clicked.connect(self.yAxisScaleChanged)
        self.canvas.mpl_connect("motion_notify_event", self.on_motion)
        self.canvas.mpl_connect("button_press_event", self.onMousePress)
        self.canvas.mpl_connect("button_release_event", self.onMouseRelease)
        self.canvas.mpl_connect("scroll_event", self.scrolling)
        self.labels = []
        self.action = None
        self.lab_lines = []
        self.MassCal = []
        self.open(filename)

    def resizeEvent(self, event):
        return
        sx, sy = self.ui.fig.width() / DPI, self.ui.fig.height() / DPI
        self.fig.set_size_inches(sx, sy)
        self.ui.fig.updateGeometry()
        self.canvas.updateGeometry()
        self.canvas.draw()
        self.canvas.flush_events()

    def __del__(self):
        if self.ita is not None:
            del self.ita
        self.ita = None

    def removeMassCalItem(self):
        row = self.ui.tableMassCal.selectedItems()[0].row()
        del self.MassCal[row]
        self.DoMassCal()

    def refresh(self):
        self.ax.get_xlim()
        self.yAxisScaleChanged()
        self.canvas.draw()
        self.canvas.flush_events()

    def next_mass(self):
        r = self.ax.get_xlim()
        self.ax.set_xlim(r[0] + 1, r[1] + 1)
        self.refresh()

    def prev_mass(self):
        r = self.ax.get_xlim()
        self.ax.set_xlim(r[0] - 1, r[1] - 1)
        self.refresh()

    def clear_labels(self):
        for x in self.labels:
            x.remove()
        self.labels[:] = []
        for x in self.lab_lines:
            self.ax.lines.remove(x)
        self.lab_lines[:] = []

    def plot_labels(self, colors=None):
        if colors is None:
            colors = ["r", "g", "b"]
        r = self.ax.get_xlim()
        E = []
        for nm in range(int(np.round(r[0], 0)), int(np.round(r[1], 0)) + 1):
            E += pySPM.utils.get_peaklist(nm, self.ita.polarity == "Negative")
        m0s = [pySPM.utils.get_mass(x) for x in E]
        P = list(zip(m0s, E))
        P.sort(key=lambda x: x[0])
        y = self.ax.get_ylim()[1]
        for i, (mi, Ei) in enumerate(P):
            dmi = (
                2
                * np.sqrt(mi)
                * np.sqrt(
                    (self.dk0**2 / (self.sf**2)) + mi * (self.dsf**2 / (self.sf**2))
                )
            )
            col = colors[i % len(colors)]
            self.lab_lines.append(self.ax.axvline(mi, color=col, alpha=0.5))
            self.labels.append(
                self.ax.annotate(
                    Ei,
                    (mi, y),
                    (5, 0),
                    rotation=90,
                    va="top",
                    ha="left",
                    textcoords="offset pixels",
                )
            )
            if dmi > 0:
                self.lab_lines.append(
                    self.ax.axvline(mi + dmi, color=col, alpha=0.5, ls=":")
                )
                self.lab_lines.append(
                    self.ax.axvline(mi - dmi, color=col, alpha=0.5, ls=":")
                )

    def yAxisScaleChanged(self):
        r = self.ax.get_xlim()
        delta = r[1] - r[0]
        self.clear_labels()

        if self.ita is not None:
            SatLevel = (
                self.ita.size["pixels"]["x"]
                * self.ita.size["pixels"]["y"]
                * self.ita.Nscan
            )
            self.sat_level.set_ydata(SatLevel)

        max = 0
        left = int(pySPM.utils.mass2time(r[0], sf=self.sf, k0=self.k0) / 2)
        right = int(pySPM.utils.mass2time(r[1], sf=self.sf, k0=self.k0) / 2) + 1
        if left < 0:
            left = 0
        if right >= len(self.S):
            right = len(self.S) - 1
        if left < self.t[-1] and right > 0:
            max = np.max(self.S[left : right + 1])
            if max > 0:
                self.ax.set_ylim(0, 1.2 * max)
        if delta < 10:
            self.ui.show_mass.setEnabled(True)
            if self.ui.show_mass.isChecked():
                self.plot_labels()
        else:
            self.ui.show_mass.setEnabled(False)
        m0 = pySPM.utils.time2mass(left + right, self.sf, self.k0)
        dm = (
            2
            * np.sqrt(m0)
            * np.sqrt((self.dk0**2 / (self.sf**2)) + m0 * (self.dsf**2 / (self.sf**2)))
        )
        self.ui.lab_m0.setText(f"m0 = {m0:.5f} ± {dm:.5f}")

    def scrolling(self, event):
        r = self.ax.get_xlim()
        r[1] - r[0]
        m0 = event.xdata
        zfact = 2
        if event.button == "down":
            zfact = 1 / zfact
        low = m0 - (m0 - r[0]) * zfact
        high = m0 + (r[1] - m0) * zfact
        self.ax.set_xlim((low, high))
        self.refresh()

    def open(self, t_filename=None):
        settings = QSettings(QSettings.IniFormat, QSettings.UserScope, "pySPM", "pySPM")
        if t_filename is None:
            home = QDir.cleanPath(os.getenv("HOMEPATH"))
            path = settings.value("lastPath", home)
            self.filename = QFileDialog.getOpenFileName(
                None, "Choose measurement file", path, "*.ita"
            )[0]
        else:
            self.filename = t_filename
        check_file = QFileInfo(self.filename)
        self.setWindowTitle(check_file.fileName())
        if not check_file.exists() or not check_file.isFile():
            return

        settings.setValue("lastPath", check_file.path())
        self.ita = pySPM.ITA(self.filename, readonly=False)
        self.t, self.S = self.ita.getSpectrum(time=True)
        self.sf, self.k0 = self.ita.get_mass_cal()
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.spec = self.ax.plot(self.mass, self.S)[0]
        SatLevel = (
            self.ita.size["pixels"]["x"] * self.ita.size["pixels"]["y"] * self.ita.Nscan
        )
        self.sat_level = self.ax.axhline(SatLevel, color="r")
        if "pySPM" in self.ita.root.goto("MassScale"):
            self.MassCal = []
            N = self.ita.root.goto("MassScale/pySPM/N").get_ulong()
            for i in range(N):
                elt = self.ita.root.goto(
                    "MassScale/pySPM/" + str(i) + "/elt"
                ).value.decode("utf8")
                mass = self.ita.root.goto(
                    "MassScale/pySPM/" + str(i) + "/mass"
                ).get_double()
                time = self.ita.root.goto(
                    "MassScale/pySPM/" + str(i) + "/time"
                ).get_double()
                self.MassCal.append({"elt": elt, "mass": mass, "time": time})
        else:
            self.MassCal = []
            for x in self.ita.root.goto("MassScale/calib"):
                if x.name == "assign":
                    self.MassCal.append({"elt": x.get_string()})
                if x.name == "mcp":
                    mcp = struct.unpack("<10d", x.value)
                    self.MassCal[-1]["time"] = mcp[0]
                    self.MassCal[-1]["mass"] = mcp[1]
        self.DoMassCal()

    def get_mass(self, formula):
        if self.ita is not None and not (
            formula.endsWith("+") or formula.endsWith("-")
        ):
            pol = "-" if self.ita.polarity == "Negative" else "+"
            formula = formula + pol
        return pySPM.utils.get_mass(formula)

    def plotSpectra(self):
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.spec.set_xdata(self.mass)
        self.refresh()

    def onMousePress(self, event):
        if event.button == 3:
            x = event.xdata

            i = pySPM.utils.closest_arg(self.mass, x)
            last = i - 1
            while i != last:
                last = i
                i = i - 10 + np.argmax(self.S[i - 10 : i + 10])
            self.S[i]
            self.MassCal.append({"time": self.t[i]})
            print(f"clicked @{x}u (t={self.t[i]})")
        elif event.button == Qt.LeftButton:
            self.action = ("move", event.xdata)
        else:
            print(event)

    def on_motion(self, event):
        r = self.ax.get_xlim()
        if self.action is not None and self.action[0] == "move":
            if event.xdata is not None:
                r[1] - r[0]
                dx = event.xdata - self.action[1]
                self.ax.set_xlim((r[0] - dx, r[1] - dx))
                self.refresh()

    def onMouseRelease(self, event):
        if event.button == 3:
            x = event.xdata
            r = self.ax.get_xlim()
            frags = []
            for nm in range(int(np.round(r[0], 0)), int(np.round(r[1], 0)) + 1):
                frags += pySPM.utils.get_peaklist(nm, self.ita.polarity == "Negative")
            masses = np.array([pySPM.utils.get_mass(x) for x in frags])
            dm = masses - x
            i = np.argmin(np.abs(dm))
            self.MassCal[-1]["mass"] = masses[i]
            self.MassCal[-1]["elt"] = frags[i]
            print(f"assigned to {frags[i]}")
            self.DoMassCal()
        self.action = None
        self.refresh()

    def save_mass_cal(self):
        root = self.ita.root.goto("MassScale")
        for i, mcp in enumerate(self.MassCal):
            for key in mcp:
                if key in ["elt"]:
                    data = mcp[key].encode("utf8") + b"\x00" * (256 - len(mcp[key]))
                    root.edit_block("pySPM/" + str(i), key, data)
                else:
                    root.edit_block("pySPM/" + str(i), key, struct.pack("<d", mcp[key]))
        root.edit_block("pySPM", "N", struct.pack("<I", len(self.MassCal)))

    def DoMassCal(self):
        ts = [x["time"] for x in self.MassCal]
        ms = [x["mass"] for x in self.MassCal]
        if len(ts) > 1:
            self.sf, self.k0, self.dsf, self.dk0 = pySPM.utils.fitSpectrum(
                ts, ms, error=True
            )
        else:
            self.k0 = ts[0] - self.sf * np.sqrt(ms[0])
        self.ui.lab_k0.setText(f"k0 = {self.k0} ± {self.dk0}")
        self.ui.lab_sf.setText(f"sf = {self.sf} ± {self.dsf}")

        self.ita.setK0(self.k0)
        self.ita.setSF(self.sf)

        self.ui.tableMassCal.clearContents()
        self.ui.tableMassCal.setRowCount(len(self.MassCal))
        for i in range(len(self.MassCal)):
            self.ui.tableMassCal.setItem(i, 0, QTableWidgetItem(self.MassCal[i]["elt"]))
            m = pySPM.utils.time2mass(self.MassCal[i]["time"], self.sf, self.k0)
            self.ui.tableMassCal.setItem(
                i, 1, QTableWidgetItem("{:.3f}".format(self.MassCal[i]["mass"]))
            )
            self.ui.tableMassCal.setItem(
                i, 2, QTableWidgetItem("{:.0f}".format(self.MassCal[i]["time"]))
            )
            delta = "{:.6f}".format(m - self.MassCal[i]["mass"])
            self.ui.tableMassCal.setItem(i, 3, QTableWidgetItem(delta))
        self.mass = pySPM.utils.time2mass(self.t, self.sf, self.k0)
        self.save_mass_cal()
        self.plotSpectra()

    def toggleMassCal(self):
        vis = not self.ui.tableMassCal.isVisible()
        self.ui.tableMassCal.setVisible(vis)
        self.ui.pushButton.setVisible(vis)
        if vis:
            self.ui.pushButton_2.setText("«")
        else:
            self.ui.pushButton_2.setText("»")

    def closeEvent(self, event):
        print("Closing spectra tool...")
        del self.ita


def main(filename=None):
    if filename is None and len(sys.argv) > 1:
        filename = sys.argv[1]
    print(f'Loading file "{filename}"')
    app = QApplication(sys.argv)
    window = SpectraViewer(filename)
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import sys

cmd = None

if len(sys.argv) > 1:
    cmd = sys.argv[1]

commands = {}


def run_script(func):
    import importlib

    mod = importlib.import_module("pySPM.tools." + func)
    sys.argv = [sys.argv[0], *list(sys.argv[2:])]
    mod.main()


cmds = {
    "stability": {
        "doc": "Run a stability plotter tool in a GUI",
        "script": "stability",
    },
    "plotter": {
        "doc": "Display a live graphical display of all the logged parameters (usually the emission current)",
        "alias": ["emission_current_plotter", "parameters_plotter"],
        "script": "emission_current_plotter",
    },
    "timer": {
        "doc": "Display informations about the current measurement time. In particular this function estimate the remaining time from the measurement proportion and the elapsed time.",
        "alias": ["tof_timer", "measurement_timer"],
        "script": "tof_timer",
    },
    "spectra": {"doc": "Spectra plotter and mass calibration GUI", "script": "spectra"},
}

for c in cmds:
    doc = cmds[c].get("doc", "")
    script = cmds[c].get("script", None)
    commands[c] = (script, doc)
    if "alias" in cmds[c]:
        for a in cmds[c]["alias"]:
            commands[a] = (script, doc)

if cmd is None or cmd in ["help", "--help", "-h", "?", "-?"]:
    print("Give as argument the command that pySPM should run. Valid commands are:")
    for c in cmds:
        print(
            "{}: {} (aliases: {})".format(
                c, cmds[c].get("doc", ""), ",".join(cmds[c].get("alias", []))
            )
        )
elif cmd in commands:
    print(f'Running command "{cmd}"')
    run_script(commands[cmd][0])
else:
    print(f'Command "{cmd}" not understood')
    print("Valid commands are:")
    for c in commands:
        print(f"{c}: {commands[c][1]}")



# Form implementation generated from reading ui file 'spectraviewer.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtWidgets

from .mplwidget import MplWidget


class Ui_SpectraViewer:
    def setupUi(self, SpectraViewer):
        SpectraViewer.setObjectName("SpectraViewer")
        SpectraViewer.resize(1148, 690)
        self.centralwidget = QtWidgets.QWidget(SpectraViewer)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.pushButton_2 = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_2.sizePolicy().hasHeightForWidth())
        self.pushButton_2.setSizePolicy(sizePolicy)
        self.pushButton_2.setMaximumSize(QtCore.QSize(18, 16777215))
        self.pushButton_2.setObjectName("pushButton_2")
        self.gridLayout.addWidget(self.pushButton_2, 1, 2, 4, 1)
        self.lab_sf = QtWidgets.QLabel(self.centralwidget)
        self.lab_sf.setObjectName("lab_sf")
        self.gridLayout.addWidget(self.lab_sf, 2, 0, 1, 2)
        self.tableMassCal = QtWidgets.QTableWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tableMassCal.sizePolicy().hasHeightForWidth())
        self.tableMassCal.setSizePolicy(sizePolicy)
        self.tableMassCal.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableMassCal.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableMassCal.setColumnCount(4)
        self.tableMassCal.setObjectName("tableMassCal")
        self.tableMassCal.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.tableMassCal.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableMassCal.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableMassCal.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.tableMassCal.setHorizontalHeaderItem(3, item)
        self.tableMassCal.horizontalHeader().setDefaultSectionSize(64)
        self.gridLayout.addWidget(self.tableMassCal, 1, 0, 1, 2)
        self.lab_k0 = QtWidgets.QLabel(self.centralwidget)
        self.lab_k0.setObjectName("lab_k0")
        self.gridLayout.addWidget(self.lab_k0, 3, 0, 1, 2)
        self.pushButton = QtWidgets.QPushButton(self.centralwidget)
        self.pushButton.setObjectName("pushButton")
        self.gridLayout.addWidget(self.pushButton, 4, 0, 1, 2)
        self.mpl = MplWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mpl.sizePolicy().hasHeightForWidth())
        self.mpl.setSizePolicy(sizePolicy)
        self.mpl.setStyleSheet("")
        self.mpl.setObjectName("mpl")
        self.gridLayout.addWidget(self.mpl, 1, 3, 4, 1)
        self.lab_m0 = QtWidgets.QLabel(self.centralwidget)
        self.lab_m0.setObjectName("lab_m0")
        self.gridLayout.addWidget(self.lab_m0, 0, 3, 1, 1)
        self.show_mass = QtWidgets.QCheckBox(self.centralwidget)
        self.show_mass.setEnabled(False)
        self.show_mass.setCheckable(True)
        self.show_mass.setChecked(True)
        self.show_mass.setObjectName("show_mass")
        self.gridLayout.addWidget(self.show_mass, 0, 0, 1, 2)
        SpectraViewer.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(SpectraViewer)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1148, 21))
        self.menubar.setObjectName("menubar")
        SpectraViewer.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(SpectraViewer)
        self.statusbar.setObjectName("statusbar")
        SpectraViewer.setStatusBar(self.statusbar)

        self.retranslateUi(SpectraViewer)
        QtCore.QMetaObject.connectSlotsByName(SpectraViewer)

    def retranslateUi(self, SpectraViewer):
        _translate = QtCore.QCoreApplication.translate
        SpectraViewer.setWindowTitle(_translate("SpectraViewer", "MainWindow"))
        self.pushButton_2.setText(_translate("SpectraViewer", "«"))
        self.lab_sf.setText(_translate("SpectraViewer", "sf"))
        item = self.tableMassCal.horizontalHeaderItem(0)
        item.setText(_translate("SpectraViewer", "Element"))
        item = self.tableMassCal.horizontalHeaderItem(1)
        item.setText(_translate("SpectraViewer", "Mass"))
        item = self.tableMassCal.horizontalHeaderItem(2)
        item.setText(_translate("SpectraViewer", "Time"))
        item = self.tableMassCal.horizontalHeaderItem(3)
        item.setText(_translate("SpectraViewer", "Delta [u]"))
        self.lab_k0.setText(_translate("SpectraViewer", "k0"))
        self.pushButton.setText(_translate("SpectraViewer", "Delete Element"))
        self.lab_m0.setText(_translate("SpectraViewer", "Center Mass:"))
        self.show_mass.setText(_translate("SpectraViewer", "Show masses"))



import psutil

from .win32_helper import (
    click,
    clickButton,
    findControl,
    findControls,
    findTopWindow,
    findWindow,
    getBBox,
    getPID,
    selectHoriz,
    selectVert,
)


class Fpanel:
    def __init__(self):
        self.PID = getPID("fpanel5.exe")
        if self.PID is not None:
            self.proc = psutil.Process(self.PID)
            self.hwnd = findWindow("Fpanel")[0]
        else:
            raise Exception("Fpanel is not running")

    def isInstrumentPanelOpen(self):
        try:
            hwnd = findTopWindow(wantedClass="InstrumentPanel.36575160")
            return hwnd
        except:
            return False

    def getLogFile(self):
        logs = []
        for f in self.proc.open_files():
            if (
                f.path.startswith("C:\\Program")
                or f.path.startswith("C:\\Users")
                or f.path.startswith("C:\\Windows")
            ):
                continue
            if f.path.endswith(".txt") or f.path.endswith(".log"):
                logs.append(f.path)
        if len(logs) == 0:
            log = None
        elif len(logs) == 1:
            log = logs[0]
        else:
            logs2 = [x for x in logs if "log" in x]
            log = logs2[0] if len(logs) == 1 else logs[0]
        return log

    def openInstrumentPanel(self):
        controls = []
        for x in findControls(self.hwnd, wantedClass="Button"):
            bbox = getBBox(x)
            w = bbox[2] - bbox[0]
            h = bbox[3] - bbox[1]
            if w != 28 or h != 28:
                continue
            controls.append([x, *bbox])
        Y = sorted({x[2] for x in controls})
        hInstrumentButton = next(x[0] for x in controls if x[2] == Y[3])
        clickButton(hInstrumentButton)
        clickButton(hInstrumentButton)

    def openTab(self, i):
        if not self.isInstrumentPanelOpen():
            self.openInstrumentPanel()
        hwnd = findTopWindow(wantedClass="InstrumentPanel.36575160")
        hwnd = selectVert(selectHoriz(findControls(hwnd, wantedClass="Button"), 0), i)[
            0
        ]
        clickButton(hwnd)
        clickButton(hwnd)

    def openLogWindow(self):
        if not self.isLogWindowOpen():
            self.openTab(8)
            hwnd = findTopWindow(wantedClass="InstrumentPanel.36575160")
            hBt = findControl(hwnd, wantedText="Log...", wantedClass="Button")
            clickButton(hBt)
            clickButton(hBt)
        return findTopWindow(wantedClass="LogFrame.259931064")

    def isLogWindowOpen(self):
        hwnd = findWindow("Log Settings")
        if len(hwnd) == 0:
            return False
        return hwnd[0]

    def populateLogFields(self):
        hlog = self.openLogWindow()
        hwnd = findTopWindow(wantedClass="InstrumentPanel.36575160")
        hwnds = findControls(hwnd, wantedClass="PythonCustomControl.Meter.36566376")
        if True:
            hAdd = findControl(hlog, "Add", "Button")
            clickButton(hAdd)
            self.openTab(1)
            hEC = selectVert(hwnds, 1)[0]
            bbox = getBBox(hEC)
            x = (bbox[0] + bbox[2]) // 2
            y = (bbox[1] + bbox[3]) // 2
            click(x, y)



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Helper functions to handle spectras.
"""

from .misc import alias, dec_debug, do_debug


def get_substance_peaks(substance, negative=True):
    import os
    import sqlite3

    DB_PATH = os.path.join(
        os.path.abspath(os.path.join(__file__, "../..")), "data", "elements.db"
    )
    conn = sqlite3.connect(DB_PATH)
    c = conn.cursor()
    c.execute(
        "SELECT Peaks.Fragment from Peaks where Peaks.Substance==(SELECT ID from substance where Name LIKE '%{name}%') and Polarity{pol}=0".format(
            name=substance, pol="><"[negative]
        )
    )
    return [x[0] for x in c.fetchall()]


def get_dm(m, sf, k0, dsf, dk0):
    import numpy as np

    return 2 * np.sqrt(m) * np.sqrt((dk0**2 / (sf**2)) + m * (dsf**2 / (sf**2)))


@alias("showPeak")
def show_peak(
    m,
    D,
    m0,
    delta=None,
    errors=False,
    dm0=0,
    dofit=False,
    show_elts=True,
    debug=False,
    Aredux=1,
    label=None,
    include=None,
    include_only=None,
    exclude=None,
    polarity="+",
    colors="rgb",
    pretty=True,
    formula=True,
    auto_scale=True,
    fakefit=False,
    zero_axis=True,
    **kargs,
):
    """
    Plot a spectrum (given by its mass vector m and intensity vector D) around a mass m0 (Δm=delta).

    Parameters
    ----------
    m : numpy.ndarray
        mass vector
    D : numpy.ndarray
        Intensity vector
    m0 : float
        Central mass of the plot's x-axis
    delta: None or float
        the half-span of the plot's x-axis.
        If None will try to determine automatically the peaks region
    errors : boolean
        If true will include the fitting errors (if dofit is True)
    dofit : boolean
        If True will fit the peaks for the elements found
    show_elts : boolean
        If True will display a list of all elements for the plotted region. The basic elements comes from a database, but they can be adjusted with the include, include_only and exclude parameters.
    include : list
        List of elements to be included to the plotting elements
    include_only : list
        Only those elements will be displayed and none from the database
    exclude : list
        Exclude some elements from display
    fakefit : boolean
        If the fit fails, fakefit can be used in order to see and adjust the first fitting guess
    dm0 : float
        The mass shift used for the fitting first guess.
    Aredux : float
        The amplitude reduction used to the first fitting guess.
    label : string
        Plot label. Will be displayed if you use the legend() function from matplotlib.
    polarity : string ("+" or "-")
        will switch the elements chosen from the database
    colors : string
        A series of letters defining matplotlib color used to display the elements.
    pretty : boolean
        If True will display the elements in a pretty way such that the elements text are not overlapping (or at least will avoid it as much as possible)
    formula : boolean
        Will convert the formula in a pretty manner. Exponents and indices will be written correctly
    auto_scale : boolean
        If True will try to rescale the plot in order to have all labels inside.
    zero_axis : boolean or dict
        Will draw an horizontal line for the zero intensities.
        If a dict, the parameters of the lines can be tuned. See the parameters passed to axhline from matplotlib
    """
    import copy

    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.optimize import curve_fit

    from . import LG, get_mass, get_peaklist
    from .elts import formulafy

    if exclude is None:
        exclude = []
    if "showElts" in kargs:
        from warnings import warn

        warn("Parameter showElts is deprecated. Please use show_elts", stacklevel=2)
        show_elts = kargs.pop("showElts")

    if include is None:
        include = []

    if isinstance(include, str):
        include = include.split(",")

    assert isinstance(include, list)

    if isinstance(m0, str):
        include.append(m0)
        m0 = get_mass(m0)

    if do_debug(debug):
        import time

        time.time()

    if isinstance(exclude, str):
        exclude = exclude.split(",")

    if fakefit:
        dofit = True

    if delta is None:
        delta0 = kargs.pop("delta0", 0.5)
        mask = (m >= m0 - delta0) * (m <= m0 + delta0)
        thresh = max(5, np.max(D[mask]) * kargs.get("min_perc", 0.01))
        li = np.argmax(mask) + np.argmax(D[mask] > thresh)
        ui = np.argmax(mask) + np.sum(mask) - np.argmax(D[mask][::-1] > thresh)
        upper_mass = m[ui] + 0.01
        lower_mass = m[li] - 0.01

        delta = (upper_mass - lower_mass) / 2.0 + 0.01
    else:
        lower_mass = m0 - delta
        upper_mass = m0 + delta
    if do_debug(debug):
        print(f"Lower mass: {lower_mass:.2f}\nUpper mass: {upper_mass:.2f}")
    mask = (m >= lower_mass) * (m <= upper_mass)

    negative = False
    if polarity in ["-", "Negative", "negative", "neg", "Neg", "NEG"]:
        negative = True
    if show_elts:
        if include_only is not None:
            E = (
                include_only.split(",")
                if isinstance(include_only, str)
                else include_only
            )
        else:
            E = [
                x
                for NM in range(int(round(lower_mass)), int(round(upper_mass)) + 1)
                for x in get_peaklist(NM, negative)
            ]
            E = [x for x in E if x not in exclude] + include
        E = list({x + [["+", "-"][negative], ""]["+" in x or "-" in x] for x in E})
    else:
        E = []
    if negative:
        E = [x + ["-", ""]["-" in x] for x in E]
    if do_debug(debug):
        print("Selected Elements: " + ", ".join(E))
    m0s = [get_mass(x) for x in E]
    E = [x for x, y in zip(E, m0s) if y >= lower_mass and y <= upper_mass]
    m0s = [get_mass(x) for x in E]
    E_labels = [formulafy(x) for x in E] if formula else E
    if do_debug(debug):
        print("Displayed elements:", ", ".join(E))

    Dt = np.copy(D[mask])
    mt = m[mask]
    if do_debug(debug):
        print(" ; ".join(E))
    ax = kargs.pop("ax", plt.gca())
    if dofit or fakefit:
        mp = m[mask][np.argmax(D[mask])]  # mass of max peak
        dm = dm0
        if dm is None:
            dm = 0
        if len(E) > 0:
            i = np.argmin(
                abs(
                    np.array(
                        [get_mass(x) for x in E if isinstance(x, str)]
                        + [x for x in E if isinstance(x, float)]
                    )
                    - mp
                )
            )  # which element is the closest to max_peak
            if dm0 is None:
                dm = mp - get_mass(E[i])
        p0 = [kargs.get("asym0", 1), dm] + [0, 0] * len(
            E
        )  # delta m is estimated from the deviation of the highest peak
        Et = copy.deepcopy(E)  # copy element list

        while len(Et) > 0:
            mp = mt[np.argmax(Dt)]  # mass of max peak
            ms = [get_mass(x) for x in Et]
            i = np.argmin(abs(ms - mp))
            idx = E.index(Et[i])
            j = np.argmin(abs(mt - ms[i] - dm))
            if do_debug(debug):
                print("Max element:", Et[i], mp, ms[i], Dt[j], idx)
            p0[2 + 2 * idx] = kargs.get("sig0", 0.002)
            p0[3 + 2 * idx] = Aredux * Dt[j]
            Dt -= LG(mt, ms[i], p0[2 + 2 * idx], amp=p0[3 + 2 * idx], asym=p0[0], lg=0)
            Dt[Dt < 0] = 0
            del Et[i]

        def fit(x, *p):
            y = x * 0
            for i in range((len(p) - 2) // 2):
                x0 = m0s[i] + p[1]
                y += LG(x, x0, p[2 + 2 * i], amp=p[3 + 2 * i], asym=p[0], lg=0)
            return y

        fit_type = None
        if dofit:
            try:
                assert not fakefit
                popt, pcov = curve_fit(
                    fit,
                    m[mask],
                    D[mask],
                    p0=p0,
                    bounds=(
                        [1 / kargs.get("asym_max", 10), -0.015]
                        + [0, 0] * ((len(p0) - 1) // 2),
                        [kargs.get("asym_max", 10), 0.015]
                        + [kargs.get("sig_max", 0.01), np.inf] * ((len(p0) - 1) // 2),
                    ),
                )
                fit_type = 0
            except:
                p0[1] = 0
                try:
                    assert not fakefit
                    popt, pcov = curve_fit(
                        fit,
                        m[mask],
                        D[mask],
                        p0=p0,
                        bounds=(
                            [1 / kargs.get("asym_max", 4), -0.015]
                            + [0, 0] * ((len(p0) - 1) // 2),
                            [kargs.get("asym_max", 4), 0.015]
                            + [kargs.get("sig_max", 0.01), np.inf]
                            * ((len(p0) - 1) // 2),
                        ),
                    )
                    fit_type = 1
                except Exception as e:
                    fit_type = 2
                    if do_debug(debug):
                        raise e
                    popt = p0
                    pcov = np.zeros((len(p0), len(p0)))
                    if ax is not None:
                        for x in ["right", "left", "top", "bottom"]:
                            ax.spines[x].set_color("red")
        else:
            popt = p0
            popt[1] = 0
            pcov = np.zeros((len(p0), len(p0)))
        if fakefit:
            popt[1] = dm0
        res = {}
        err = np.sqrt(np.diag(pcov))
        for i in range((len(popt) - 1) // 2):
            Y = LG(
                m[mask], m0s[i], popt[2 + 2 * i], popt[3 + 2 * i], lg=0, asym=popt[0]
            )
            Area = (
                popt[2 * i + 3]
                * popt[2 * i + 2]
                * np.sqrt(2 * np.pi)
                * (0.5 + 0.5 * popt[0])
            )
            Area_err = np.sqrt(
                (
                    err[2 * i + 3]
                    * popt[2 * i + 2]
                    * np.sqrt(2 * np.pi)
                    * (0.5 + 0.5 * popt[0])
                )
                ** 2
                + (
                    err[2 * i + 2]
                    * popt[2 * i + 3]
                    * np.sqrt(2 * np.pi)
                    * (0.5 + 0.5 * popt[0])
                )
                ** 2
                + (
                    err[0]
                    * popt[2 * i + 2]
                    * popt[2 * i + 3]
                    * np.sqrt(2 * np.pi)
                    * 0.5
                )
                ** 2
            )
            if not errors:
                res[E_labels[i]] = {
                    "m0": m0s[i],
                    "mass": m0s[i] + popt[1],
                    "Area": Area,
                    "Amp": popt[2 * i + 3],
                    "sig": popt[2 * i + 2],
                    "assym": popt[0],
                    "dm": popt[1] * 1e6 / m0s[i],
                    "fit": fit_type,
                }
            else:
                res[E_labels[i]] = {
                    "m0": m0s[i],
                    "mass": (m0s[i] + popt[1], err[1]),
                    "Area": (Area, Area_err),
                    "Amp": (popt[2 * i + 3], err[2 * i + 3]),
                    "sig": (popt[2 * i + 2], err[2 * i + 2]),
                    "assym": (popt[0], err[0]),
                    "dm": (popt[1] * 1e6 / m0s[i], err[1] * 1e6 / m0s[i]),
                    "fit": fit_type,
                }
    else:
        popt = [0, 0]
        res = None
    if ax is not None:
        if label is None:
            p = ax.plot(
                m[mask] - popt[1], D[mask], color=kargs.get("curve_color", None)
            )
        else:
            p = ax.plot(
                m[mask] - popt[1],
                D[mask],
                label=label,
                color=kargs.get("curve_color", None),
            )

    if show_elts and ax is not None:
        from . import put_Xlabels

        if dofit and ax is not None:
            ax.plot(m[mask], D[mask], color=p[0].get_color(), alpha=0.1)
            if pretty:
                put_Xlabels(
                    ax,
                    m0s,
                    [f"{E}: {res[E]['Area']:.2f}" for E in res],
                    color=colors,
                    debug=dec_debug(debug),
                    **kargs,
                )
            resO = [(res[x]["m0"], res[x]) for x in res]
            resO.sort(key=lambda x: x[0])
            for i, (_, r) in enumerate(resO):
                col = colors[i % len(colors)]
                Y = LG(m[mask], r["m0"], r["sig"], r["Amp"], lg=0, asym=r["assym"])
                ax.plot(m[mask], Y, "--", color=col)
        elif pretty:
            put_Xlabels(
                ax, m0s, E_labels, color=colors, debug=dec_debug(debug), **kargs
            )

        if "dsf" in kargs and "dk0" in kargs and "sf" in kargs and "k0" in kargs:
            sf, dsf, k0, dk0 = (kargs[x] for x in "sf,dsf,k0,dk0".split(","))
            P = list(zip(m0s, E))
            P.sort(key=lambda x: x[0])
            for i, (mi, Ei) in enumerate(P):
                col = colors[i % len(colors)]
                dmi = get_dm(mi, sf, k0, dsf, dk0)
                ax.axvline(mi - dmi, color=col, alpha=0.5, linestyle=":")
                ax.axvline(mi + dmi, color=col, alpha=0.5, linestyle=":")

        if not pretty:
            P = list(zip(m0s, E_labels))
            P.sort(key=lambda x: x[0])
            y = ax.get_ylim()[1]
            for i, (mi, Ei) in enumerate(P):
                col = colors[i % len(colors)]
                ax.axvline(mi, color=col, alpha=0.5)
                ax.annotate(
                    Ei,
                    (mi, y),
                    (5, 0),
                    rotation=90,
                    va="top",
                    ha="left",
                    textcoords="offset pixels",
                    color=col,
                )
        if auto_scale:
            fig = plt.gcf()
            renderer = fig.canvas.get_renderer()
            ymax = ax.get_ylim()[1]
            ymin = ax.get_ylim()[0]
            for child in ax.get_children():
                bbox = child.get_window_extent(renderer)
                bbox_data = bbox.transformed(ax.transData.inverted())
                if ymax < bbox_data.ymax:
                    ymax = bbox_data.ymax
            ax.set_ylim((ymin, ymax))
            ax.autoscale_view()

            # Add the Intensity=0 line in case there is negative intensities (happens when plotting PCA components)
    if zero_axis:
        if isinstance(zero_axis, dict):
            ax.achline(0, **zero_axis)
        else:
            ax.axhline(0, color="k", alpha=0.5, lw=0.5)

    if dofit and ax is not None:
        ax.plot(m[mask] - popt[1], fit(m[mask], *popt), "r:")
    return res


def plot_isotopes(
    elt,
    amp=None,
    main=None,
    ax=None,
    sig=None,
    asym=None,
    lg=0,
    limit=1,
    color="C1",
    show_elts=False,
    debug=False,
    dm=0,
    **kargs,
):
    """
    plot the isotopes of a given element on a spectral profile plot
    """
    # Old parameter name compatibility
    if "Amp" in kargs:
        from warnings import warn

        warn(
            "Parameter Amp is deprecated. Please use amp in order to set the amplitude!",
            stacklevel=2,
        )
        amp = kargs.pop("Amp")
    if "showElts" in kargs:
        from warnings import warn

        warn("Parameter showElts is deprecated. Please use show_elts", stacklevel=2)
        show_elts = kargs.pop("showElts")
    import re

    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.optimize import curve_fit

    from . import LG, Molecule, get_abund, get_isotopes, get_mass

    if type(elt) is Molecule:
        elt = str(elt)
    if ax is None:
        ax = plt.gca()
    if main is None:
        main = ax
    main_isotope = re.sub("^([0-9]+)", "", elt)
    main_iso = (main_isotope, get_mass(main_isotope), get_abund(main_isotope))
    L = main.lines[0]
    m, y = L.get_xdata(), L.get_ydata()
    m -= dm
    if hasattr(ax, "log") and ax.log:
        y = 10**y
    mask = np.abs(m - main_iso[1]) < 0.1
    if amp is None:
        i = np.argmin(np.abs(m - main_iso[1]))
        amp = np.max(y[i - 10 : i + 10]) / main_iso[2]
    if sig is None and asym is None:

        def fit(x, s, a):
            return LG(x, main_iso[1], s, amp=amp * main_iso[2], lg=lg, asym=a)

        (sig, asym), _ = curve_fit(fit, m[mask], y[mask], (0.005, 1))
    elif sig is None:

        def fit(x, s):
            return LG(x, main_iso[1], s, amp=amp * main_iso[2], lg=lg, asym=asym)

        s, _ = curve_fit(fit, m[mask], y[mask], (0.005))
        sig = s[0]
    elif asym is None:

        def fit(x, a):
            return LG(x, main_iso[1], sig, amp=amp * main_iso[2], lg=lg, asym=a)

        a, _ = curve_fit(fit, m[mask], y[mask], (1))
        asym = a[0]
    if debug:
        print(sig, asym)
    isos = get_isotopes(elt, min_abund=limit / amp)
    L = ax.lines[0]
    m, y = L.get_xdata(), L.get_ydata()
    if hasattr(ax, "log") and ax.log:
        y = 10**y
    main.get_xlim()
    s = m * 0

    for iso in isos:
        s += LG(m, iso[1], sig, amp=amp * iso[2], lg=lg, asym=asym)
        if show_elts:
            ax.annotate(
                iso[0],
                (iso[1], amp * iso[2]),
                (0, 5),
                textcoords="offset pixels",
                ha="center",
            )
    if hasattr(ax, "log") and ax.log:
        s[s >= 1] = np.log10(s[s >= 1])
        s[s < 1] = 0
    ax.plot(m + dm, s, color, **kargs)
    return m, s



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Provides useful functions for fitting with scipy.optimize.curve_fit
"""

import numpy as np

from . import math


def peak_fit(m, s, m0, delta=0.02):
    from scipy.optimize import curve_fit

    from . import LG, get_mass

    assert len(m) == len(s)
    if isinstance(m0, str):
        m0 = get_mass(m0)
    assert m0 < m[-1] and m0 > m[0]
    assert delta > 0
    mask = (m >= (m0 - delta)) * (m <= (m0 + delta))
    amp0 = np.max(s[mask])
    sig = abs(m[mask][np.argmin(abs(s[mask] - amp0 / 2))] - m0) / np.sqrt(2 * np.log(2))
    p00 = (m[mask][np.argmax(s[mask])], sig, amp0, 0.3, 1)
    p0, _ = curve_fit(
        LG,
        m[mask],
        s[mask],
        p00,
        bounds=((-np.inf, 0, 0, 0, 0), (np.inf, np.inf, np.inf, 1, np.inf)),
    )
    return p0


def CDF(x, bg, *args, **kargs):
    """
    Return the sum of several CDFs.
    x: values used to evaluate the function
    bg: background (value on the very left)
    args: suits of one or several Ai, xi, si describing each a CDF
        Ai: Amplitude
        xi: position
        si: sigma of CDF
                     _
    ex: step edge: _| |_ between 0.2 and 0.7 in y and -1 and 1 in x with sigma=5
        CDF(x, 0.2, 0.5, -1, 5, -0.5, 1, 5)
    """
    if "A" in kargs and "x0" in kargs and "sig" in kargs:
        args = [x for y in zip(kargs["A"], kargs["x0"], kargs["sig"]) for x in y]

    r = bg * np.ones(x.shape)
    if len(args) % 3 != 0:
        raise Exception("Invalid number of arguments. see help.")
        return

    for i in range(len(args) // 3):
        r += args[3 * i] * math.CDF(x, args[3 * i + 1], args[3 * i + 2])
    return r


def lgCDF(x, bg, lg=0, *args, **kargs):
    """
    Return the sum of several CDFs.
    x: values used to evaluate the function
    bg: background (value on the very left)
    lg: Lorentz-Gauss factor (0=Gauss, 1=Lorentz, in-between = mix)
    args: suits of one or several Ai, xi, si describing each a CDF
        Ai: Amplitude
        xi: position
        si: sigma of CDF
                     _
    ex: step edge: _| |_ between 0.2 and 0.7 in y and -1 and 1 in x with sigma=5
        lgCDF(x, lg, 0.2, 0.5, -1, 5, -0.5, 1, 5)
    """

    if "A" in kargs and "x0" in kargs and "sig" in kargs:
        args = [x for y in zip(kargs["A"], kargs["x0"], kargs["sig"]) for x in y]

    r = bg * np.ones(x.shape)
    if len(args) % 3 != 0:
        raise Exception("Invalid number of arguments. see help.")
        return

    for i in range(len(args) // 3):
        r += args[3 * i] * math.CDF(x, args[3 * i + 1], args[3 * i + 2], lg=lg)
    return r


def lgCDF_fit(x, y, p0, dic=False):
    import scipy.optimize as opt

    n = (len(p0) - 2) // 3
    bounds = [
        [0, 0] + [-np.inf, -np.inf, 0] * n,
        [np.inf, 1] + [np.inf, np.inf, np.inf] * n,
    ]
    popt, pcov = opt.curve_fit(lgCDF, x, y, p0, bounds=bounds)
    if dic:
        d = np.diag(pcov)
        return {
            "bg": popt[0],
            "lg": popt[1],
            "A": popt[2::3],
            "x0": popt[3::3],
            "sig": popt[4::3],
        }, {"bg": d[0], "lg": d[1], "A": d[2::3], "x0": d[3::3], "sig": d[4::3]}
    return popt, pcov


def CDF_fit(x, y, p0, dic=False):
    import scipy.optimize as opt

    n = (len(p0) - 1) // 3
    bounds = [[0] + [-np.inf, -np.inf, 0] * n, [np.inf] + [np.inf, np.inf, np.inf] * n]
    popt, pcov = opt.curve_fit(CDF, x, y, p0, bounds=bounds)
    if dic:
        d = np.diag(pcov)
        return {"bg": popt[0], "A": popt[1::3], "x0": popt[2::3], "sig": popt[3::3]}, {
            "bg": d[0],
            "A": d[2::3],
            "x0": d[3::3],
            "sig": d[4::3],
        }
    return popt, pcov


def LG2Dr(A, ratio=None, Rweight=None, sigma=None, dic=False, **kargs):
    """
    Perform a 2D Lorentz-Gauss fitting on a 2D array A with a fix ration between σx and σy
    dic: if True return the solution as a dictionary

    output: list of popt and pcov array.
    The parameter order is:
    Amplitude, Angle, Sigma_x, Sigma_y, Center_x, Center_y, LGx, LGy
    """
    if ratio is None:
        ratio = np.sqrt(2)

    from scipy.optimize import curve_fit

    params = ["amplitude", "angle", "sig_y", "x0", "y0", "LG_x", "LG_y", "bg"]

    def fit(XY, *args):
        # Add default parameters to the argument list at the right position
        j = 0
        a = []
        for x in params:
            if x in kargs:
                a.append(kargs[x])
            else:
                a.append(args[j])
                j += 1
        a = a[:2] + [ratio * a[2]] + a[2:]  # insert sig_x
        # Calculate the 2D Lorentz-Gauss
        return a[-1] + math.LG2D(XY, *a[:-1]).ravel()

    # First guess for parameters
    Amplitude = np.max(A)
    Center = np.unravel_index(np.argmax(A), A.shape)
    p0_def = [Amplitude, 0, 10, Center[1], Center[0], 0, 0, 0]
    bounds_def = [
        [0, np.radians(-45), 0, -np.inf, -np.inf, 0, 0, 0],
        [np.inf, np.radians(45), +np.inf, np.inf, np.inf, 1, 1, np.inf],
    ]
    p0 = [p0_def[i] for i, x in enumerate(params) if x not in kargs]
    bounds = [
        [bounds_def[0][i] for i, x in enumerate(params) if x not in kargs],
        [bounds_def[1][i] for i, x in enumerate(params) if x not in kargs],
    ]
    # Kick out arguments for the given default parameters

    x = np.arange(A.shape[1])
    y = np.arange(A.shape[0])
    X, Y = np.meshgrid(x, y)
    R = np.sqrt((X - Center[1]) ** 2 + (Y - Center[0]) ** 2)
    if Rweight is not None:
        sigma = 0.001 + R if Rweight == "R" else Rweight(R)
    if sigma is not None:
        sigma = np.ravel(sigma)
    pfit = curve_fit(fit, (X, Y), np.ravel(A), p0=p0, sigma=sigma, bounds=bounds)

    # Re-add default parameters to the output
    pout = []
    dpout = []
    j = 0
    for _i, x in enumerate(params):
        if x in kargs:
            pout.append(kargs[x])
            dpout.append(0)
        else:
            pout.append(pfit[0][j])
            dpout.append(np.sqrt(pfit[1][j, j]))
            j += 1
    pout = pout[:2] + [ratio * pout[2]] + pout[2:]
    dpout = dpout[:2] + [ratio * dpout[2]] + dpout[2:]
    if dic:
        params2 = params[:2] + ["sig_x"] + params[2:]
        return dict(zip(params2, pout)), dict(zip(params2, dpout))
    return pout, dpout


def LG2D(A, Rweight=None, sigma=None, dic=False, **kargs):
    """
    Perform a 2D Lorentz-Gauss fitting on a 2D array A
    dic: if True return the solution as a dictionary

    output: list of popt and pcov array.
    The parameter order is:
    Amplitude, Angle, Sigma_x, Sigma_y, Center_x, Center_y, LGx, LGy
    """
    from scipy.optimize import curve_fit

    params = [
        "amplitude",
        "angle",
        "sig_x",
        "sig_y",
        "x0",
        "y0",
        "LG_x",
        "LG_y",
        "assym_x",
        "assym_y",
        "bg",
    ]

    def fit(XY, *args):
        # Add default parameters to the argument list at the right position
        j = 0
        a = []
        for x in params:
            if x in kargs:
                a.append(kargs[x])
            else:
                a.append(args[j])
                j += 1
        # Calculate the 2D Lorentz-Gauss
        return a[-1] + math.LG2D(XY, *a[:-1]).ravel()

    # First guess for parameters
    Amplitude = np.max(A)
    Center = np.unravel_index(np.argmax(A), A.shape)
    p0_def = [Amplitude, 0, 10, 10, Center[1], Center[0], 0, 0, 1, 1, 0]
    bounds_def = [
        [0, np.radians(-45), 0, 0, -np.inf, -np.inf, 0, 0, 0, 0, -np.inf],
        [
            np.inf,
            np.radians(45),
            +np.inf,
            +np.inf,
            np.inf,
            np.inf,
            1,
            1,
            np.inf,
            np.inf,
            np.inf,
        ],
    ]
    p0 = [p0_def[i] for i, x in enumerate(params) if x not in kargs]
    bounds = [
        [bounds_def[0][i] for i, x in enumerate(params) if x not in kargs],
        [bounds_def[1][i] for i, x in enumerate(params) if x not in kargs],
    ]
    # Kick out arguments for the given default parameters

    x = np.arange(A.shape[1])
    y = np.arange(A.shape[0])
    X, Y = np.meshgrid(x, y)
    R = np.sqrt((X - Center[1]) ** 2 + (Y - Center[0]) ** 2)
    if Rweight is not None:
        sigma = 0.001 + R if Rweight == "R" else Rweight(R)
    if sigma is not None:
        sigma = np.ravel(sigma)
    pfit = curve_fit(fit, (X, Y), np.ravel(A), p0=p0, sigma=sigma, bounds=bounds)

    # Re-add default parameters to the output
    pout = []
    dpout = []
    j = 0
    for _i, x in enumerate(params):
        if x in kargs:
            pout.append(kargs[x])
            dpout.append(0)
        else:
            pout.append(pfit[0][j])
            dpout.append(np.sqrt(pfit[1][j, j]))
            j += 1

    if dic:
        return dict(zip(params, pout)), dict(zip(params, dpout))
    return pout, dpout


def LG2Da(A, Rweight=None, sigma=None, dic=False, **kargs):
    """
    Perform a 2D Lorentz-Gauss fitting on a 2D array A
    dic: if True return the solution as a dictionary

    output: list of popt and pcov array.
    The parameter order is:
    Amplitude, Angle, Sigma_x, Sigma_y, Center_x, Center_y, LGx, LGy
    """
    from scipy.optimize import curve_fit

    params = [
        "amplitude",
        "angle",
        "sigN",
        "sigS",
        "sigE",
        "sigW",
        "x0",
        "y0",
        "LGN",
        "LGS",
        "LGE",
        "LGW",
        "bg",
    ]

    def fit(XY, *args):
        # Add default parameters to the argument list at the right position
        j = 0
        a = []
        for x in params:
            if x in kargs:
                a.append(kargs[x])
            else:
                a.append(args[j])
                j += 1
        # Calculate the 2D Lorentz-Gauss
        return a[-1] + math.LG2Da(XY, *a[:-1]).ravel()

    # First guess for parameters
    Amplitude = np.max(A)
    Center = np.unravel_index(np.argmax(A), A.shape)
    #   Amplitude,           angle,    sigN,    sigS,    sigE,    sigW,        x0,        y0, LGN, LGS, LGE, LGW, bg
    p0_def = [Amplitude, 0, 10, 10, 10, 10, Center[1], Center[0], 0, 0, 0, 0, 0]
    bounds_def = [
        [0, np.radians(-45), 0, 0, 0, 0, -np.inf, -np.inf, 0, 0, 0, 0, -np.inf],
        [
            np.inf,
            np.radians(45),
            +np.inf,
            +np.inf,
            +np.inf,
            +np.inf,
            +np.inf,
            np.inf,
            1,
            1,
            1,
            1,
            np.inf,
        ],
    ]
    p0 = [p0_def[i] for i, x in enumerate(params) if x not in kargs]
    bounds = [
        [bounds_def[0][i] for i, x in enumerate(params) if x not in kargs],
        [bounds_def[1][i] for i, x in enumerate(params) if x not in kargs],
    ]
    # Kick out arguments for the given default parameters

    x = np.arange(A.shape[1])
    y = np.arange(A.shape[0])
    X, Y = np.meshgrid(x, y)
    R = np.sqrt((X - Center[1]) ** 2 + (Y - Center[0]) ** 2)
    if Rweight is not None:
        sigma = 0.001 + R if Rweight == "R" else Rweight(R)
    if sigma is not None:
        sigma = np.ravel(sigma)
    pfit = curve_fit(fit, (X, Y), np.ravel(A), p0=p0, sigma=sigma, bounds=bounds)

    # Re-add default parameters to the output
    pout = []
    dpout = []
    j = 0
    for _i, x in enumerate(params):
        if x in kargs:
            pout.append(kargs[x])
            dpout.append(0)
        else:
            pout.append(pfit[0][j])
            dpout.append(np.sqrt(pfit[1][j, j]))
            j += 1

    if dic:
        return dict(zip(params, pout)), dict(zip(params, dpout))
    return pout, dpout



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import os
import pickle
import shutil
import tempfile
import zipfile

"""
This small utility can save and load python objects to a single file.
It uses ZIP DEFLATE compression method in order to keep the file small.
It works in a similar way as numpy.savez_compressed and numpy.load, but
just work with any python objects and thus do not generate a numpy.array
for every object. Several objects can be added to the same file with either one call of save or several.
The new data will just be appended to your file. It also support the update of the file (in this case the data file is copied in a new one).

the default file extension is pkz (for Pickle Zip)
"""

data_path = "."


def set_datapath(path):
    global data_path
    data_path = path


def findPKZ(filename):
    if os.path.splitext(filename)[1] == "":
        filename += ".pkz"
    p2 = os.path.join(data_path, filename)
    if os.path.exists(p2):
        return p2
    if os.path.exists(filename):
        return filename
    if not os.path.exists(filename):
        raise OSError(f'File "{filename}" not found')
    return filename


def inarxiv(filename, obj):
    if os.path.splitext(filename)[1] == "":
        filename += ".pkz"
    filename = os.path.join(data_path, filename)
    out = zipfile.ZipFile(filename, "a", zipfile.ZIP_DEFLATED)
    file_list = out.namelist()
    return obj in file_list


def save(filename, *objs, **obj):
    """
    save python objects to file
    """
    if os.path.splitext(filename)[1] == "":
        filename += ".pkz"
    filename = os.path.join(data_path, filename)
    out = zipfile.ZipFile(filename, "a", zipfile.ZIP_DEFLATED)
    file_list = out.namelist()
    update = []
    for i, o in enumerate(objs):
        obj[i] = o

    # List objects which are already present in the file (so which will be updated)
    for k in obj:
        if k in file_list:
            update.append(k)

    if len(update) == 0:  # No files are updated, just append the new data to the file
        for k in obj:
            out.writestr(k, pickle.dumps(obj[k], pickle.HIGHEST_PROTOCOL))
    else:
        # close the current file, copy it to a temp file and reopen the original file in write mode (will thus override all data)
        out.close()
        ft, temp = tempfile.mkstemp()
        shutil.copy(filename, temp)
        out = zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED)

        # Copy all old data which are not updated and clean tempfile
        old = zipfile.ZipFile(temp, "r")
        for k in [x for x in file_list if x not in update]:
            out.writestr(k, old.read(k))
        old.close()
        os.fdopen(ft).close()
        os.remove(temp)

        # Write all new objects
        for k in obj:
            out.writestr(k, pickle.dumps(obj[k], pickle.HIGHEST_PROTOCOL))
    out.close()


def load(filename, *keys):
    """
    load python objects from saved pkz files
    """
    filename = findPKZ(filename)
    if len(keys) == 0:
        keys = ["0"]
    elif len(keys) == 1 and "," in keys[0]:
        keys = keys[0].split(",")
    f = zipfile.ZipFile(filename, "r")

    res = []
    for key in keys:
        if isinstance(key, int):
            key = str(key)
        try:
            raw = f.read(key)
        except Exception:
            raise KeyError(f"There is no key {key} found in the data {filename}")
        try:
            res.append(pickle.loads(raw))
        except:
            raise Exception(f"Cannot pickle recorded data for key {key}")

    f.close()
    if len(keys) == 1:
        return res[0]
    return res


class loader:
    """
    This class act as a dictionary and read default value for compressed pkz files.
    Values can be set to new values without changing the content of the saved data.
    Only retrieved data are kept in memory and not all the data.
    """

    def __init__(self, filename):
        self.filename = findPKZ(filename)
        self.local = {}

    def __iter__(self):
        f = zipfile.ZipFile(self.filename, "r")
        self.keys = set(f.namelist() + list(self.local.keys()))
        f.close()
        return self

    def __next__(self):
        if len(self.keys) == 0:
            raise StopIteration()
        else:
            return self.keys.pop()

    def __getitem__(self, key):
        if key not in self.local:
            f = zipfile.ZipFile(self.filename, "r")
            self.local[key] = pickle.loads(f.read(key))
            f.close()
        return self.local[key]

    def __setitem__(self, key, value):
        self.local[key] = value

    def __delitem__(self, key):
        del self.local[key]


class BidirData:
    """
    This class act as a dictionary and read default value for compressed pkz files.
    Values can be set to new values without changing the content of the saved data.
    Only retrieved data are kept in memory and not all the data.
    """

    def __init__(self, filename):
        try:
            self.filename = findPKZ(filename)
        except OSError:
            if os.path.splitext(filename)[1] == "":
                filename += ".pkz"
            self.filename = os.path.join(data_path, filename)
            out = zipfile.ZipFile(filename, "a", zipfile.ZIP_DEFLATED)
            out.close()
        self.local = {}

    def __iter__(self):
        f = zipfile.ZipFile(self.filename, "r")
        self.keys = set(f.namelist() + list(self.local.keys()))
        f.close()
        return self

    def __next__(self):
        if len(self.keys) == 0:
            raise StopIteration()
        else:
            return self.keys.pop()

    def __getitem__(self, key):
        if key not in self.local:
            f = zipfile.ZipFile(self.filename, "r")
            self.local[key] = pickle.loads(f.read(key))
            f.close()
        return self.local[key]

    def __setitem__(self, key, value):
        self.local[key] = value
        save(self.filename, **{key: value})

    def __delitem__(self, key):
        del self.local[key]

    def keys(self):
        f = zipfile.ZipFile(self.filename, "r")
        keys = f.filelist
        f.close()
        return [x.filename for x in keys]
